@@ -67,6 +67,10 @@
  * PWS 1996/12/10
  */
 
+#ifdef HAVE_STDIO_H
+#include <stdio.h>
+#endif
+
 #include "zsh.mdh"
 #include "input.pro"
 
@@ -218,16 +222,21 @@ ingetc(void)
 static int
 inputline(void)
 {
-    char *ingetcline, *ingetcpmptl = NULL, *ingetcpmptr = NULL;
+    char *ingetcline, **ingetcpmptl = NULL, **ingetcpmptr = NULL;
+    int context = ZLCON_LINE_START;
 
     /* If reading code interactively, work out the prompts. */
     if (interact && isset(SHINSTDIN)) {
-	if (!isfirstln)
-	    ingetcpmptl = prompt2;
+	if (!isfirstln) {
+	    ingetcpmptl = &prompt2;
+	    if (rprompt2)
+		ingetcpmptr = &rprompt2;
+	    context = ZLCON_LINE_CONT;
+	}
 	else {
-	    ingetcpmptl = prompt;
+	    ingetcpmptl = &prompt;
 	    if (rprompt)
-		ingetcpmptr = rprompt;
+		ingetcpmptr = &rprompt;
 	}
     }
     if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))) {
@@ -246,7 +255,8 @@ inputline(void)
 	     */
 	    char *pptbuf;
 	    int pptlen;
-	    pptbuf = unmetafy(promptexpand(ingetcpmptl, 0, NULL, NULL), &pptlen);
+	    pptbuf = unmetafy(promptexpand(ingetcpmptl ? *ingetcpmptl : NULL,
+					   0, NULL, NULL), &pptlen);
 	    write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
 	    free(pptbuf);
 	}
@@ -262,8 +272,10 @@ inputline(void)
 	 * typeahead when the terminal settings are altered.
 	 *                     pws 1998/03/12
 	 */
-	ingetcline = (char *)zleread(ingetcpmptl, ingetcpmptr,
-				     ZLRF_HISTORY|ZLRF_NOSETTY);
+	int flags = ZLRF_HISTORY|ZLRF_NOSETTY;
+	if (isset(IGNOREEOF))
+	    flags |= ZLRF_IGNOREEOF;
+	ingetcline = zlereadptr(ingetcpmptl, ingetcpmptr, flags, context);
 	histdone |= HISTFLAG_SETTY;
     }
     if (!ingetcline) {
@@ -273,9 +285,6 @@ inputline(void)
 	free(ingetcline);
 	return lexstop = errflag = 1;
     }
-    /* Look for a space, to see if this shouldn't be put into history */
-    if (isfirstln)
-	spaceflag = *ingetcline == ' ';
     if (isset(VERBOSE)) {
 	/* Output the whole line read so far. */
 	zputs(ingetcline, stderr);
@@ -374,7 +383,7 @@ inungetc(int c)
 	     * can't back up where we want to.  Instead, we just push it
 	     * onto the input stack as an extra character.
 	     */
-	    char *cback = (char *)zcalloc(2);
+	    char *cback = (char *)zshcalloc(2);
 	    cback[0] = (char) c;
 	    inpush(cback, INP_FREE|INP_CONT, NULL);
 	}
@@ -410,7 +419,7 @@ stuff(char *fn)
     off_t len;
 
     if (!(in = fopen(unmeta(fn), "r"))) {
-	zerr("can't open %s", fn, 0);
+	zerr("can't open %s", fn);
 	return 1;
     }
     fseek(in, 0, 2);
@@ -418,7 +427,7 @@ stuff(char *fn)
     fseek(in, 0, 0);
     buf = (char *)zalloc(len + 1);
     if (!(fread(buf, len, 1, in))) {
-	zerr("read error on %s", fn, 0);
+	zerr("read error on %s", fn);
 	fclose(in);
 	zfree(buf, len + 1);
 	return 1;
@@ -543,3 +552,29 @@ inpop(void)
 	inpoptop();
     } while (remcont);
 }
+
+/*
+ * Expunge any aliases from the input stack; they shouldn't appear
+ * in the history and need to be flushed explicitly when we encounter
+ * an error.
+ */
+
+/**/
+void
+inpopalias(void)
+{
+    while (inbufflags & INP_ALIAS)
+	inpoptop();
+}
+
+
+/*
+ * Get pointer to remaining string to read.
+ */
+
+/**/
+char *
+ingetptr(void)
+{
+    return inbufptr;
+}
